@@ -22,6 +22,17 @@ def get_at_time(df, t, time_label = 'Time'):
                 dft[i] = sliced.iloc[0]
     return pd.DataFrame(dft).T
 
+def Sobol_sequence(lbd, ubd, power_no):
+    """
+    Create 2^power_no of inputs for sampling based on the lists of lbd (lower
+    bound) and ubd (upper bound).
+    """
+    from scipy.stats import qmc
+    sampler = qmc.Sobol(d = len(lbd), scramble = False)
+    inputs = sampler.random_base2(m = power_no)
+    inputs = qmc.scale(inputs, lbd, ubd)
+    return inputs
+
 def update_ds(ds):
     """
     Updates old ds options with the new ones
@@ -52,10 +63,10 @@ def qp(df, constraints, vnames, x = None, opt = None):
     from dside import DSI
     
     ds = DSI(df)
-    p = ds.screen(constraints)
-    r = ds.plot(vnames, opt)
+    ds.screen(constraints)
+    ds.plot(vnames, opt)
     if x != None:
-        r = ds.find_AOR(x)
+        ds.find_AOR(x)
     return ds
 
 class DSI():
@@ -82,7 +93,7 @@ class DSI():
         }
         self.all_x = {}
         self.space_size = None
-        self.tetra = None
+        self.DTspreadsheet = None
         self.r = None
         
         # Set default color map for heat map plot
@@ -196,10 +207,11 @@ class DSI():
             'extra_points': [], # Extra points to be used to design space identification
             'maxiter': 50,    # Maximum bisection iterations
             'tol': 0.001,     # Bisection tolerance (on the MV which is amul)
-            'lb': 0.001,      # Lower bound of initial bisection run
-            'ub': 5,          # Upper bound of initial bisection run
-            'printF': True,  # If true, print iter details
-            'maxvp': 0,       # Maximum allowed percentage of vio in DSp
+            'lb': 0.001,        # Lower bound of initial bisection run
+            'ub': 5,            # Upper bound of initial bisection run
+            'printF': True,     # If true, print iter details
+            'maxvp': 0,         # Maximum allowed percentage of vio in DSp
+            'check_tol': 1e-20, # Point check calculation error tolerance
         }
         self.opt = self.default_opt.copy()
         self.bw_template = {'alpha': 1, 'satcolor': 'gray', 'viocolor': 'black', \
@@ -242,6 +254,12 @@ class DSI():
         vio['SatFlag'] = False
         self.sat = sat
         self.vio = vio
+        
+        # Update saved
+        self.df['SatFlag'] = 'N/A'
+        self.df.loc[sat.index, 'SatFlag'] = True
+        self.df.loc[vio.index, 'SatFlag'] = False
+        self.df['SatFlag'].astype('bool')
         return None
     
     def help(self, print_opt = False):
@@ -397,18 +415,34 @@ class DSI():
                 self.find_DSp(vnames, opt = opt)
             shp = self.shp
             if dim == 2:
-                plt.plot(*zip(*shp['edges_val']),\
-                    linewidth = opt['dspwidth'], linestyle = opt['dspstyle'],\
-                    color = opt['dspcolor'], label = opt['dsplabel'],\
-                    zorder = opt['dspzorder'])
+                for i in range(len(shp['reg_bounds_val'])):
+                    if i == 0:
+                        plt.plot(*zip(*shp['reg_bounds_val'][i]),\
+                            linewidth = opt['dspwidth'], linestyle = opt['dspstyle'],\
+                            color = opt['dspcolor'], label = opt['dsplabel'],\
+                            zorder = opt['dspzorder'])
+                    else:
+                        plt.plot(*zip(*shp['reg_bounds_val'][i]),\
+                            linewidth = opt['dspwidth'], linestyle = opt['dspstyle'],\
+                            color = opt['dspcolor'],\
+                            zorder = opt['dspzorder'])
 
             if dim == 3:
-                surf = ax.plot_trisurf(*zip(*shp['P']), triangles = shp['tri'], \
-                    color = opt['dspcolor'],\
-                    alpha = opt['dspalpha'], label = opt['dsplabel'],\
-                    zorder = opt['dspzorder'])
-                surf._facecolors2d=surf._facecolor3d
-                surf._edgecolors2d=surf._edgecolor3d
+                for i in range(len(shp['reg_bounds'])):
+                    if i == 0:
+                        surf = ax.plot_trisurf(*zip(*shp['P']), triangles = shp['reg_bounds'][i], \
+                            color = opt['dspcolor'],\
+                            alpha = opt['dspalpha'], label = opt['dsplabel'],\
+                            zorder = opt['dspzorder'])
+                        surf._facecolors2d=surf._facecolor3d
+                        surf._edgecolors2d=surf._edgecolor3d
+                    else:
+                        surf = ax.plot_trisurf(*zip(*shp['P']), triangles = shp['reg_bounds'][i], \
+                            color = opt['dspcolor'],\
+                            alpha = opt['dspalpha'],\
+                            zorder = opt['dspzorder'])
+                        surf._facecolors2d=surf._facecolor3d
+                        surf._edgecolors2d=surf._edgecolor3d
         
         # Limits for axes
         if opt['axeslimdf'] != 'best':
@@ -546,20 +580,41 @@ class DSI():
         if print_flag:
             print(sol_flag)
 
-        # Calculate volume
+        # Calculate design space size and classify regions
         if dim == 3:
+            shp = self.classify_regions3D(shp)
             volume = 0
-            for v in shp['P'][shp['tetras']]:
+            for v in shp['P'][shp['simplices']]:
                 volume += np.abs(np.dot(v[0] - v[3], np.cross(v[1] - v[3], v[2] - v[3])))/6
             shp['size'] = volume
-        self.vindsp = vio[inside(vpoints, shp)]
+        else:
+            shp = self.classify_regions2D(shp)
+            P = shp['P']
+            simps = shp['simplices']
+            v = simps
+            a = (P[v][:, 0][:, 0] - P[v][:, 1][:, 0])**2 + \
+                (P[v][:, 0][:, 1] - P[v][:, 1][:, 1])**2
+            b = (P[v][:, 1][:, 0] - P[v][:, 2][:, 0])**2 + \
+                (P[v][:, 1][:, 1] - P[v][:, 2][:, 1])**2
+            c = (P[v][:, 2][:, 0] - P[v][:, 0][:, 0])**2 + \
+                (P[v][:, 2][:, 1] - P[v][:, 0][:, 1])**2
+            a, b, c = np.sqrt([a, b, c])
+            s = (a + b + c)*0.5
+            area = np.sqrt(s*(s - a)*(s - b)*(s - c)) # area from Heron's formula
+            shp['size'] = np.sum(area)
+    
+        if vpoints.shape[0] == 0:
+            self.vindsp = None
+        else:
+            self.vindsp = vio[inside(vpoints, shp)]
         self.indsp = pd.concat([sat, self.vindsp]).reset_index()
         space_size = shp['size']
         self.shp = shp
         self.space_size = space_size
-        self.bF = shp['tri']
+        self.bF = shp['simplices']
         self.P = shp['P']
         self.report['space_size'] = space_size
+        self.report['no_reg'] = self.shp['no_reg']
         self.report['sol_flag'] = sol_flag
         # self.report['opt_log'] = opt_log
         
@@ -572,163 +627,409 @@ class DSI():
         """
         Calculate the alphashape boundary from a point cloud P (2D np array)
         """
-        from shapely.geometry import MultiLineString
-        from shapely.ops import unary_union, polygonize
         from scipy.spatial import Delaunay
-        from collections import Counter
         import numpy as np
-        import itertools
+        import pandas as pd
 
-        v = Delaunay(P).vertices
-        # Calculate the sides of the triangles
-        a = (P[v][:, 0][:, 0] - P[v][:, 1][:, 0])**2 + \
-            (P[v][:, 0][:, 1] - P[v][:, 1][:, 1])**2
-        b = (P[v][:, 1][:, 0] - P[v][:, 2][:, 0])**2 + \
-            (P[v][:, 1][:, 1] - P[v][:, 2][:, 1])**2
-        c = (P[v][:, 2][:, 0] - P[v][:, 0][:, 0])**2 + \
-            (P[v][:, 2][:, 1] - P[v][:, 0][:, 1])**2
-        a, b, c = np.sqrt([a, b, c])
-        s = (a + b + c)*0.5
-        area = np.sqrt(s*(s - a)*(s - b)*(s - c)) # area from Heron's formula
-        alpha_filter = a*b*c/(4.0*area) < alpha
+        spreadsheet = self.DTspreadsheet
+        if type(spreadsheet) == type(None):
+            # Get Delaunay triangulation of the points
+            tri = Delaunay(P)
+            simps = tri.simplices
+            triP = P[simps]
 
-        # Filter vertices of alpha shape based on alpha radius
-        tri = v[alpha_filter]
-        # edges = np.vstack([edges[:, 0:2], edges[:, 1:3], np.delete(edges, 2, 1)])
-        # edges = list(zip(edges[:, 0], edges[:, 1]))
-        edges = [tuple(sorted(cb)) for e in tri for cb in itertools.combinations(e, 2)]
+            # --- Calculating circumcircle radius and center (vectorised) --- #
+            PA = triP[:, 0, :]; Ax = PA[:, 0]; Ay = PA[:, 1]
+            PB = triP[:, 1, :]; Bx = PB[:, 0]; By = PB[:, 1]
+            PC = triP[:, 2, :]; Cx = PC[:, 0]; Cy = PC[:, 1]
 
-        count = Counter(edges)
-        # Keep only edges that appear one time (concave hull edges)
-        edges = [e for e, c in count.items() if c == 1]
-        edges_val = [(P[e[0]], P[e[1]]) for e in edges]
+            # translation of vertex A to the origin
+            Bxo = Bx - Ax; Cxo = Cx - Ax # Axo = Ax - Ax; 
+            Byo = By - Ay; Cyo = Cy - Ay # Ayo = Ay - Ay; 
 
-        # Return points in order for plotting
-        ml = MultiLineString(edges_val)
-        poly = polygonize(ml)
-        hull = unary_union(list(poly))
-        hull_vertices = hull.exterior.coords.xy
-        h = np.zeros((len(hull_vertices[0]), 2))
-        h[:, 0] = hull_vertices[0]
-        h[:, 1] = hull_vertices[1]
+            Do = 2*(Bxo*Cyo - Byo*Cxo)
+            Uxo = (1/Do)*(Cyo*(Bxo**2 + Byo**2) - Byo*(Cxo**2 + Cyo**2))
+            Uyo = (1/Do)*(Bxo*(Cxo**2 + Cyo**2) - Cxo*(Bxo**2 + Byo**2))
+            R = np.sqrt(Uxo**2 + Uyo**2)          # radius of circumcircle
+            U = np.array([Uxo + Ax, Uyo + Ay]).T  # center of circumcircle
 
-        shp = {}
-        shp['P'] = P
-        shp['verts'] = np.unique(edges)
-        shp['tri'] = tri
-        shp['tetras'] = None
-        shp['edges'] = edges
-        shp['edges_val'] = h
-        shp['alpha'] = alpha
+            spreadsheet = pd.DataFrame(simps, columns = ['p1', 'p2', 'p3'])
+            spreadsheet[['p1x', 'p1y']] = PA
+            spreadsheet[['p2x', 'p2y']] = PB
+            spreadsheet[['p3x', 'p3y']] = PC
+            spreadsheet[['Ux', 'Uy']] = U
+            spreadsheet['r'] = R
 
-        # Area of alpha shape
-        v = tri
-        a = (P[v][:, 0][:, 0] - P[v][:, 1][:, 0])**2 + \
-            (P[v][:, 0][:, 1] - P[v][:, 1][:, 1])**2
-        b = (P[v][:, 1][:, 0] - P[v][:, 2][:, 0])**2 + \
-            (P[v][:, 1][:, 1] - P[v][:, 2][:, 1])**2
-        c = (P[v][:, 2][:, 0] - P[v][:, 0][:, 0])**2 + \
-            (P[v][:, 2][:, 1] - P[v][:, 0][:, 1])**2
-        a, b, c = np.sqrt([a, b, c])
+            self.DTspreadsheet = spreadsheet
 
-        s = (a + b + c)*0.5
-        area = np.sqrt(s*(s - a)*(s - b)*(s - c)) # area from Heron's formula
-        shp['size'] = np.sum(area)
+        alpha_spreadsheet = spreadsheet[spreadsheet['r'] <= alpha].copy()
+
+        # ----- Get edge lines only ----- #
+        simps = alpha_spreadsheet[['p1', 'p2', 'p3']].to_numpy()
+        edgeComb = np.array([(0, 1), (0, 2), (1, 2)]) # comb to separate tri to lines
+        edges = simps[:, edgeComb].reshape(-1, 2) # separate tri to lines
+        edges.sort(axis = 1) # sort vertex indices
+        edges = pd.DataFrame(edges).drop_duplicates(keep = False).to_numpy() # get unique entries only
+
+        shp = {
+            'P': P, 
+            'simplices': simps, 
+            'edges': edges, 
+            'alpha': alpha,
+            'alpha_spreadsheet': alpha_spreadsheet,
+            }
+        return shp
+
+    
+    def classify_regions2D(self, shp = None):
+        """
+        Classify number of regions and their outer lines
+        """
+        import numpy as np
+        import pandas as pd
+
+        # Unpack shp
+        if type(shp) == None:
+            shp = self.shp
+        edges = shp['edges']
+        P = shp['P']
+
+        # ----- Identification of Regions and Ordering Boundary ----- #
+        # Implemented 2D breadth-first search
+        # Create worksheet
+        ws = pd.DataFrame(edges[edges[:, 0].argsort()])
+        ws['visit'] = False
+        ws['region'] = np.NaN
+
+        change_init_flag = False
+        reg = 0
+        reg_bounds = [] # record ordered boundary vertices
+        # Loop over every line to categorise them into regions
+        while ws[ws['visit'] == False].shape[0] > 0:
+            reg += 1
+
+            # create sorted bounds
+            bounds = []
+            # find a point which have not been visited yet
+            cur_idx = ws[ws['visit'] == False].index[0]
+            ws.loc[cur_idx, 'visit'] = True # set it to be visited
+            ws.loc[cur_idx, 'region'] = reg # categorise to region reg
+            
+            # taking both directions of the line to search neighbouring vertices
+            current_vert1 = int(ws.loc[cur_idx][0])
+            current_vert2 = int(ws.loc[cur_idx][1])
+
+            # record boundary
+            bounds.append(current_vert1)
+            bounds.append(current_vert2)
+
+            # setting flags for exploration of neighbouring vertices
+            empty_v1 = False
+            empty_v2 = False
+            change_init_flag = False
+            while change_init_flag == False: # while v1 and v2 are not empty
+                # make working_sheet for easier manipulation
+                w_s = ws[ws['visit'] == False].copy()
+
+                # get v1 and v2 (contains current_vert1 and 2)
+                v1 = pd.concat(
+                    [w_s[w_s[0] == current_vert1], 
+                     w_s[w_s[1] == current_vert1]]
+                    )[[0, 1]]
+                v2 = pd.concat(
+                    [w_s[w_s[0] == current_vert2], 
+                     w_s[w_s[1] == current_vert2]]
+                    )[[0, 1]]
+
+                # check if v1 is empty
+                if v1.shape[0] != 0:
+                    ws.loc[v1.index[0], 'visit'] = True # set it to be visited
+                    ws.loc[v1.index[0], 'region'] = reg # categorise to region reg
+                    next_vert1 = v1.T[v1.T != current_vert1].dropna().iloc[0, 0] # get next vertex
+                    current_vert1 = int(next_vert1)
+                    bounds.insert(0, current_vert1) # record boundary
+                else:
+                    empty_v1 = True
+
+                # check if v2 is empty
+                if v2.shape[0] != 0:
+                    ws.loc[v2.index[0], 'visit'] = True # set it to be visited
+                    ws.loc[v2.index[0], 'region'] = reg # categorise to region reg
+                    next_vert2 = v2.T[v2.T != current_vert2].dropna().iloc[0, 0] # get next vertex
+                    current_vert2 = int(next_vert2)
+                    bounds.append(current_vert2) # record boundary
+                else:
+                    empty_v2 = True
+                
+                # if both v1 and v2 are empty, set change_init_flag to true
+                if empty_v2 == True:
+                    if empty_v1 == True:
+                        change_init_flag = True
+                    else:
+                        pass
+                else:
+                    pass
+            # Check if the last entry in bounds is the same as first
+            if bounds[-1] != bounds[0]:
+                if bounds[-2] == bounds[0]: # check if second last is same as first
+                    bounds = bounds[:-1]    # if yes, slice boundary
+                else: # else, there may be something wrong, print warning
+                    print('WARNING: BOUNDARY FORMED MAY NOT BE CLOSED')
+            reg_bounds.append(np.array(bounds)) # compile boundary defined
+
+        # Bounds based on regions
+        ws['region'].astype('int')
+        reg_bounds_val = [P[i] for i in reg_bounds]
+
+        shp.update(
+            {
+            'ws': ws,
+            'reg_bounds': reg_bounds, 
+            'reg_bounds_val': reg_bounds_val,
+            'no_reg': len(reg_bounds),
+            }
+        )
         return shp
 
     def alphashape_3D(self, P, alpha):
         """
         Calculate the alphashape boundary from a point cloud P (3D np array)
-
         Compute the alpha shape (concave hull) of a set of 3D points.
-        Parameters:
-            pos - np.array of shape (n,3) points.
-            alpha - alpha value.
-        return
-            outer surface vertex indices, edge indices, and triangle indices
         """
         from scipy.spatial import Delaunay
         import numpy as np
         import pandas as pd
-        tetra = self.tetra
-        r = self.r
-        if tetra == None:
+
+        spreadsheet = self.DTspreadsheet
+        if type(spreadsheet) == type(None):
+            # Get Delaunay triangulation of the points
             tetra = Delaunay(P)
-            # Find radius of the circumsphere.
-            # By definition, radius of the sphere fitting inside the
-            # tetrahedral needs to be smaller than alpha value
-            # http://mathworld.wolfram.com/Circumsphere.html
+            simps = tetra.simplices
+            tetraP = P[simps]
 
-            tetrapos = np.take(P, tetra.vertices,axis=0)
-            normsq = np.sum(tetrapos**2, axis=2)[:, :, None]
-            ones = np.ones((tetrapos.shape[0], tetrapos.shape[1], 1))
+            # --- Calculating circumsphere radius and center (vectorised) --- #
+            normsq = np.sum(tetraP**2, axis=2)[:, :, None]
+            ones = np.ones((tetraP.shape[0], tetraP.shape[1], 1))
 
-            a  =  np.linalg.det(np.concatenate((tetrapos, ones), axis=2))
-            a[a == 0] = 1e-30
-            Dx =  np.linalg.det(np.concatenate((normsq, tetrapos[:,:,[1,2]], ones), axis = 2))
-            Dy = -np.linalg.det(np.concatenate((normsq, tetrapos[:,:,[0,2]], ones), axis = 2))
-            Dz =  np.linalg.det(np.concatenate((normsq, tetrapos[:,:,[0,1]], ones), axis = 2))
-            c  =  np.linalg.det(np.concatenate((normsq, tetrapos), axis = 2))
-            r  =  np.sqrt(Dx**2 + Dy**2 + Dz**2 - 4*a*c)/(2*np.abs(a))
-            self.tetra = tetra
-            self.r = r
-        # Find tetrahedrals
-        tetras = tetra.vertices[r<alpha, :]
+            A = np.linalg.det(np.concatenate((tetraP, ones), axis=2))
+            A[A == 0] = 1e-30
+            Dx = np.linalg.det(np.concatenate((normsq, tetraP[:,:,[1,2]], ones), axis = 2))
+            Dy = -np.linalg.det(np.concatenate((normsq, tetraP[:,:,[0,2]], ones), axis = 2))
+            Dz = np.linalg.det(np.concatenate((normsq, tetraP[:,:,[0,1]], ones), axis = 2))
+            C = np.linalg.det(np.concatenate((normsq, tetraP), axis = 2))
 
-        # triangles
-        triComb = np.array([(0, 1, 2), (0, 1, 3), (0, 2, 3), (1, 2, 3)])
-        triangles = tetras[:, triComb].reshape(-1, 3)
-        triangles = np.sort(triangles, axis = 1)
+            Ux = Dx/(2*A)
+            Uy = Dy/(2*A)
+            Uz = Dz/(2*A)
+            U = np.array([Ux, Uy, Uz]).T   # center of circumsphere
+            R = np.sqrt(Dx**2 + Dy**2 + Dz**2 - 4*A*C)/(2*np.abs(A)) # radius of circumsphere
+            
+            spreadsheet = pd.DataFrame(simps, columns = ['p1', 'p2', 'p3', 'p4'])
+            spreadsheet[['Ux', 'Uy', 'Uz']] = U
+            spreadsheet['r'] = R
 
-        # Remove triangles that occurs twice, because they are within shapes
-        triangles = pd.DataFrame(triangles).drop_duplicates(keep = False).to_numpy()
+            self.DTspreadsheet = spreadsheet
+            
+        alpha_spreadsheet = spreadsheet[spreadsheet['r'] <= alpha].copy()
 
-        #edges
-        edgeComb = np.array([(0, 1), (0, 2), (1, 2)])
-        edges = triangles[:, edgeComb].reshape(-1, 2)
-        edges = np.sort(edges, axis = 1)
-        edges = np.unique(edges, axis = 0)
+        # ----- Get edge triangles only ----- #
+        simps = alpha_spreadsheet[['p1', 'p2', 'p3', 'p4']].to_numpy()
+        edgeComb = np.array([(0, 1, 2), (0, 1, 3), (0, 2, 3), (1, 2, 3)]) # comb to separate tetrahedrons to triangles
+        edges = simps[:, edgeComb].reshape(-1, 3) # separate tetra to tri
+        edges.sort(axis = 1) # sort vertex indices
+        edges = pd.DataFrame(edges).drop_duplicates(keep = False).to_numpy() # get unique entries only
 
-        vertices = np.unique(edges)
+        shp = {
+            'P': P, 
+            'simplices': simps, 
+            'edges': edges, 
+            'alpha': alpha,
+            'alpha_spreadsheet': alpha_spreadsheet,
+            }
+        return shp
 
-        shp = {}
-        shp['P'] = P
-        shp['verts'] = vertices
-        shp['tri'] = triangles
-        shp['edges'] = edges
-        shp['tetras'] = tetras
-        shp['edges_val'] = None
-        shp['alpha'] = alpha
+    def categorise3D(self, ws, cur_idx, reg):
+        """
+        Recursive function to categorise triangles into regions
+        ws: worksheet is a pandas dataframe containing all info
+        cur_idx: current index in ws of the tri being checked (int)
+        reg: region number (int)
+        returns nothing, mutates the original ws dataframe
+        """
+        import pandas as pd
+        ws.loc[cur_idx, 'visit'] = True # set it to be visited
+        ws.loc[cur_idx, 'region'] = reg # categorise to region reg
 
-        # volume = 0
-        # for v in P[tetras]:
-        #     volume += np.abs(np.dot(v[0] - v[3], np.cross(v[1] - v[3], v[2] - v[3])))/6
-        shp['size'] = 420e-10
+        # taking all directions to search neighbouring vertices
+        current_pair1 = ws.loc[cur_idx]['p1']
+        current_pair2 = ws.loc[cur_idx]['p2']
+        current_pair3 = ws.loc[cur_idx]['p3']
+        
+        # make working_sheet for easier manipulation
+        w_s = ws[ws['visit'] == False].copy()
+
+        v1 = pd.concat([w_s[w_s['p1'] == current_pair1], 
+                        w_s[w_s['p2'] == current_pair1], 
+                        w_s[w_s['p3'] == current_pair1]])[['p1', 'p2', 'p3']]
+        v2 = pd.concat([w_s[w_s['p1'] == current_pair2], 
+                        w_s[w_s['p2'] == current_pair2], 
+                        w_s[w_s['p3'] == current_pair2]])[['p1', 'p2', 'p3']]
+        v3 = pd.concat([w_s[w_s['p1'] == current_pair3], 
+                        w_s[w_s['p2'] == current_pair3], 
+                        w_s[w_s['p3'] == current_pair3]])[['p1', 'p2', 'p3']]
+
+        if v1.shape[0] != 0:
+            ws.loc[v1.index[0], 'visit'] = True # set it to be visited
+            ws.loc[v1.index[0], 'region'] = reg # categorise to region reg
+            self.categorise3D(ws, v1.index[0], reg)
+        else:
+            pass
+
+        if v2.shape[0] != 0:
+            ws.loc[v2.index[0], 'visit'] = True # set it to be visited
+            ws.loc[v2.index[0], 'region'] = reg # categorise to region reg
+            self.categorise3D(ws, v2.index[0], reg)
+        else:
+            pass
+
+        if v3.shape[0] != 0:
+            ws.loc[v3.index[0], 'visit'] = True # set it to be visited
+            ws.loc[v3.index[0], 'region'] = reg # categorise to region reg
+            self.categorise3D(ws, v3.index[0], reg)
+        else:
+            pass
+        return None
+
+    def classify_regions3D(self, shp = None):
+        """
+        Classify number of regions and their outer triangle surfaces
+        """
+        import numpy as np
+        import pandas as pd
+
+        # Unpack shp
+        if type(shp) == None:
+            shp = self.shp
+        edges = shp['edges']
+        P = shp['P']
+
+        # ----- Identification of Regions ----- #
+        # Implemented 3D breadth-first search with recursive categorisation
+        sorted_tri = edges[edges[:, 0].argsort()]
+        triComb = np.array([(0, 1), (0, 2), (1, 2)]) # comb to separate tri to lines
+
+        # Create worksheet
+        ws = pd.DataFrame(sorted_tri)
+        ws['visit'] = False
+        ws['region'] = np.NaN
+
+        # Connected triangles must share at least one pair of vertices
+        ws['p1'] = [str(i) for i in sorted_tri[:, triComb][:, 0, :]]
+        ws['p2'] = [str(i) for i in sorted_tri[:, triComb][:, 1, :]]
+        ws['p3'] = [str(i) for i in sorted_tri[:, triComb][:, 2, :]]
+
+        reg = 0
+        # Loop over every triangle to categorise them into regions
+        while ws[ws['visit'] == False].shape[0] > 0:
+            reg += 1
+
+            # find a point which have not been visited yet
+            cur_idx = ws[ws['visit'] == False].index[0]
+            self.categorise3D(ws, cur_idx, reg)
+
+        # Bounds based on regions
+        ws['region'].astype('int')
+        reg_bounds = [ws[ws['region'] == (i + 1)][[0, 1, 2]].to_numpy() for i in range(int(ws['region'].max()))]
+        reg_bounds_val = [P[i] for i in reg_bounds]
+
+        shp.update(
+            {
+            'ws': ws,
+            'reg_bounds': reg_bounds, 
+            'reg_bounds_val': reg_bounds_val,
+            'no_reg': len(reg_bounds),
+            }
+        )
         return shp
 
     def inside2D(self, x, shp = None):
         """
+        Point in triangle checks based on Barycentric coordinates vectorised
+        Brute force check on all simplices in shp
+        Splits option included to handle memory limitations
         Returns False if x is not in self.shp, True otherwise (2D)
         x: list of [x, y], or [[x1, y1], [x2, y2], ...]
         """
         import numpy as np
-        from shapely.geometry import Point, Polygon
         if shp == None:
             shp = self.shp
-        poly = Polygon(shp['edges_val'])
         dim = len(np.array(x).shape)
 
-        if dim == 1:
-            x = [x]
-        vP = [Point(i[0], i[1]) for i in x]
-        out = [poly.contains(p) for p in vP]
+        no_x = np.array(x).shape[0]
+        if dim == 2:
+            no_splits = self.opt['no_splits']
+            if no_x < no_splits:
+                no_splits = int(no_x/2)
+                x_list = np.array_split(x, no_splits)
+            else:
+                x_list = np.array_split(x, no_splits)
+        else:
+            x_list = [np.array(x)]
+
+        # Find which triangle the point lies in
+        P = shp['P']
+        simps = shp['simplices']
+
+        res_list = []
+        bool_list = []
+        for X in x_list:
+            # Convert all vertices to Barycentric coordinates wrt v0
+            v0 = P[simps[:, 0],:]
+            v1 = P[simps[:, 1],:] - v0
+            v2 = P[simps[:, 2],:] - v0
+
+            # Compute transformation matrix mat
+            n_tri = len(simps)
+            v1r = v1.T.reshape((2, 1, n_tri))
+            v2r = v2.T.reshape((2, 1, n_tri))
+            mat = np.concatenate((v1r, v2r), axis = 1)
+
+            # Get inverse of transformation matrix for each triangle
+            inv_mat = np.linalg.inv(mat.T).T
+
+            # Assemble vectors for vectorised calculation wrt length of X
+            if X.size == 2:
+                X = X.reshape((1,2))
+            n_X = X.shape[0]
+            v0p = np.repeat(v0[:,:,np.newaxis], n_X, axis = 2)
+
+            # Transform X based on the origin of each local tetrahedral coordinate system
+            transX = np.einsum('imk,kmj->kij', inv_mat, X.T - v0p)
+            
+            # Perform point check:
+            #     All transformed coordinates has to be between 0 and 1
+            #     and the sum of the coordinates is less than or equal to 1
+            val = np.all(transX >=0, axis = 1) & np.all(transX <= 1, axis = 1) & (np.sum(transX, axis = 1) <= 1)
+            id_tet, id_X = np.nonzero(val) # get indices of satisfied conditions
+            res = -np.ones(n_X, dtype = id_tet.dtype) # setup results array
+                                                        # assuming -1 initially for all entries
+            res[id_X] = id_tet # if the point lies inside, then replace the -1 values with id_tet
+            
+            # Recording
+            res_list.append(res)
+            bool_list.append(res != -1)
+        res = np.concatenate(res_list)
+        bool_res = np.concatenate(bool_list)
 
         if dim == 1:
-            out = out[0]
-        return out
+            return bool_res[0]
+        return bool_res
     
     def inside3D(self, x, shp = None):
         """
+        Point in tetrahedron checks based on Barycentric coordinates vectorised
+        Brute force check on all simplices in shp
+        Splits option included to handle memory limitations
         Returns False if x is not in self.shp, True otherwise (3D)
         x: list of [x, y, z], or [[x1, y1, z1], [x2, y2, z2], ...]
         """
@@ -742,59 +1043,62 @@ class DSI():
             no_splits = self.opt['no_splits']
             if no_x < no_splits:
                 no_splits = int(no_x/2)
-            x_list = np.array_split(x, no_splits)
+                x_list = np.array_split(x, no_splits)
+            else:
+                x_list = np.array_split(x, no_splits)
         else:
-            x_list = [x]
+            x_list = [np.array(x)]
 
         # Find which tetrahedron the point lies in
-        node_coordinates = shp['P']
-        node_ids = shp['tetras']
+        P = shp['P']
+        simps = shp['simplices']
 
         res_list = []
-        for x_s in x_list:
-            p = np.array(x_s)
+        bool_list = []
+        for X in x_list:
+            # Convert all vertices to Barycentric coordinates wrt v0
+            v0 = P[simps[:, 0],:]
+            v1 = P[simps[:, 1],:] - v0
+            v2 = P[simps[:, 2],:] - v0
+            v3 = P[simps[:, 3],:] - v0
 
-            ori = node_coordinates[node_ids[:, 0],:]
-            v1 = node_coordinates[node_ids[:, 1],:] - ori
-            v2 = node_coordinates[node_ids[:, 2],:] - ori
-            v3 = node_coordinates[node_ids[:, 3],:] - ori
-            n_tet = len(node_ids)
+            # Compute transformation matrix mat
+            n_tet = len(simps)
             v1r = v1.T.reshape((3, 1, n_tet))
             v2r = v2.T.reshape((3, 1, n_tet))
             v3r = v3.T.reshape((3, 1, n_tet))
-            mat = np.concatenate((v1r, v2r, v3r), axis=1)
+            mat = np.concatenate((v1r, v2r, v3r), axis = 1)
+
+            # Get inverse of transformation matrix for each tetrahedron
             inv_mat = np.linalg.inv(mat.T).T
-            if p.size == 3:
-                p = p.reshape((1,3))
-            n_p = p.shape[0]
-            orir = np.repeat(ori[:,:,np.newaxis], n_p, axis=2)
-            newp = np.einsum('imk,kmj->kij',inv_mat,p.T-orir)
-            val = np.all(newp>=0, axis=1) & np.all(newp <=1, axis=1) & (np.sum(newp, axis=1)<=1)
-            id_tet, id_p = np.nonzero(val)
-            res = -np.ones(n_p, dtype=id_tet.dtype) # Sentinel value
-            res[id_p] = id_tet
+
+            # Assemble vectors for vectorised calculation wrt length of X
+            if X.size == 3:
+                X = X.reshape((1, 3))
+            n_X = X.shape[0]
+            v0p = np.repeat(v0[:,:,np.newaxis], n_X, axis = 2)
+
+            # Transform X based on the origin of each local tetrahedral coordinate system
+            transX = np.einsum('imk,kmj->kij', inv_mat, X.T - v0p)
+            
+            # Perform point check:
+            #     All transformed coordinates has to be between 0 and 1
+            #     and the sum of the coordinates is less than or equal to 1
+            val = np.all(transX >=0, axis = 1) & np.all(transX <= 1, axis = 1) & (np.sum(transX, axis = 1) <= 1)
+            id_tet, id_X = np.nonzero(val) # get indices of satisfied conditions
+            res = -np.ones(n_X, dtype = id_tet.dtype) # setup results array
+                                                       # assuming -1 initially for all entries
+            res[id_X] = id_tet # if the point lies inside, then replace the -1 values with id_tet
+            
+            # Recording
             res_list.append(res)
+            bool_list.append(res != -1)
         res = np.concatenate(res_list)
+        bool_res = np.concatenate(bool_list)
 
         if dim == 1:
-            x = [x]
-        # return res
-        out = []
-        for i, r in enumerate(res):
-            V = shp['P'][shp['tetras'][r]]
-            p = x[i]
-            # Find the transform matrix from orthogonal to tetrahedron system
-            v1 = V[1]-V[0] ; v2 = V[2]-V[0] ; v3 = V[3]-V[0]
-            mat = np.array((v1,v2,v3)).T
-            # mat is 3x3 here
-            M1 = np.linalg.inv(mat)
-            # apply the transform to P (v1 is the origin)
-            newp = M1.dot(p - V[0])
-            # perform test
-            out.append(np.all(newp>=0) and np.all(newp <=1) and np.sum(newp)<=1)
-        if dim == 1:
-            out = out[0]
-        return out
+            return bool_res[0]
+        return bool_res
 
     def check_point(self, x):
         """
@@ -1004,6 +1308,7 @@ class DSI():
         # DSI results
         f.write(f'\n# -------------------------- RESULTS -------------------------- #\n')
         f.write(f'Design space size: {rp["space_size"]:10f}\n')
+        f.write(f'Number of regions: {rp["no_reg"]:10d}\n')
         f.write(f'Number of samples in DSp: {self.sat.shape[0]}\n\n')
         f.write(f'Average {rp["hmv"]["name"]}:    {rp["hmv"]["mean"]:10}\n')
         f.write(f'DS Maximum {rp["hmv"]["name"]}: {rp["hmv"]["max"]:10}\n')
